from asyncio import get_running_loop

from haiway import MissingContext, ScopeMetrics, State, ctx
from pytest import mark, raises


class ExampleState(State):
    state: str = "default"


class FakeException(Exception):
    pass


@mark.asyncio
async def test_state_is_available_according_to_context():
    with raises(MissingContext):
        assert ctx.state(ExampleState).state == "default"

    async with ctx.scope("default"):
        assert ctx.state(ExampleState).state == "default"

        async with ctx.scope("specified", ExampleState(state="specified")):
            assert ctx.state(ExampleState).state == "specified"

            async with ctx.scope("modified", ExampleState(state="modified")):
                assert ctx.state(ExampleState).state == "modified"

            assert ctx.state(ExampleState).state == "specified"

        assert ctx.state(ExampleState).state == "default"

    with raises(MissingContext):
        assert ctx.state(ExampleState).state == "default"


@mark.asyncio
async def test_state_update_updates_local_context():
    with raises(MissingContext):
        assert ctx.state(ExampleState).state == "default"

    async with ctx.scope("default"):
        assert ctx.state(ExampleState).state == "default"

        with ctx.updated(ExampleState(state="updated")):
            assert ctx.state(ExampleState).state == "updated"

            with ctx.updated(ExampleState(state="modified")):
                assert ctx.state(ExampleState).state == "modified"

            assert ctx.state(ExampleState).state == "updated"

        assert ctx.state(ExampleState).state == "default"

    with raises(MissingContext):
        assert ctx.state(ExampleState).state == "default"


@mark.asyncio
async def test_exceptions_are_propagated():
    with raises(FakeException):
        async with ctx.scope("outer"):
            async with ctx.scope("inner"):
                raise FakeException()


@mark.asyncio
async def test_completions_are_called_according_to_context_exits():
    completion_future = get_running_loop().create_future()
    nested_completion_future = get_running_loop().create_future()
    executions: int = 0

    def completion(metrics: ScopeMetrics):
        nonlocal executions
        executions += 1
        completion_future.set_result(())

    def nested_completion(metrics: ScopeMetrics):
        nonlocal executions
        executions += 1
        nested_completion_future.set_result(())

    async with ctx.scope("outer", completion=completion):
        assert executions == 0

        async with ctx.scope("inner", completion=nested_completion):
            assert executions == 0

        await nested_completion_future
        assert executions == 1

    await completion_future
    assert executions == 2


@mark.asyncio
async def test_async_completions_are_called_according_to_context_exits():
    completion_future = get_running_loop().create_future()
    nested_completion_future = get_running_loop().create_future()
    executions: int = 0

    async def completion(metrics: ScopeMetrics):
        nonlocal executions
        executions += 1
        completion_future.set_result(())

    async def nested_completion(metrics: ScopeMetrics):
        nonlocal executions
        executions += 1
        nested_completion_future.set_result(())

    async with ctx.scope("outer", completion=completion):
        assert executions == 0

        async with ctx.scope("inner", completion=nested_completion):
            assert executions == 0

        await nested_completion_future
        assert executions == 1

    await completion_future
    assert executions == 2


@mark.asyncio
async def test_metrics_are_recorded_within_context():
    completion_future = get_running_loop().create_future()
    nested_completion_future = get_running_loop().create_future()
    metric: ExampleState = ExampleState()
    nested_metric: ExampleState = ExampleState()

    async def completion(metrics: ScopeMetrics):
        nonlocal metric
        metric = metrics.read(ExampleState, default=ExampleState())
        completion_future.set_result(())

    async def nested_completion(metrics: ScopeMetrics):
        nonlocal nested_metric
        nested_metric = metrics.read(ExampleState, default=ExampleState())
        nested_completion_future.set_result(())

    async with ctx.scope("outer", completion=completion):
        ctx.record(ExampleState(state="outer-in"))

        async with ctx.scope("inner", completion=nested_completion):
            ctx.record(ExampleState(state="inner"))

        await nested_completion_future
        assert nested_metric.state == "inner"

        ctx.record(
            ExampleState(state="-out"),
            merge=lambda lhs, rhs: ExampleState(state=lhs.state + rhs.state),
        )

    await completion_future
    assert metric.state == "outer-in-out"



from asyncio import CancelledError, get_running_loop, sleep
from collections.abc import AsyncGenerator, AsyncIterator

from haiway import ctx
from haiway.context.metrics import ScopeMetrics
from haiway.state.structure import State
from pytest import mark, raises


class FakeException(Exception):
    pass


@mark.asyncio
async def test_fails_when_generator_fails():
    async def generator(value: int) -> AsyncGenerator[int, None]:
        yield value
        raise FakeException()

    elements: int = 0
    with raises(FakeException):
        async for _ in ctx.stream(generator, 42):
            elements += 1

    assert elements == 1


@mark.asyncio
async def test_cancels_when_iteration_cancels():
    async def generator(value: int) -> AsyncGenerator[int, None]:
        await sleep(0)
        yield value

    elements: int = 0
    with raises(CancelledError):
        ctx.cancel()
        async for _ in ctx.stream(generator, 42):
            elements += 1

    assert elements == 0


@mark.asyncio
async def test_ends_when_generator_ends():
    async def generator(value: int) -> AsyncGenerator[int, None]:
        yield value

    elements: int = 0
    async for _ in ctx.stream(generator, 42):
        elements += 1

    assert elements == 1


@mark.asyncio
async def test_delivers_updates_when_generating():
    async def generator(value: int) -> AsyncGenerator[int, None]:
        for i in range(0, value):
            yield i

    elements: list[int] = []

    async for element in ctx.stream(generator, 10):
        elements.append(element)

    assert elements == list(range(0, 10))


@mark.asyncio
async def test_streaming_context_variables_access_is_preserved():
    class TestState(State):
        value: int = 42
        other: str = "other"

    async def generator(value: int) -> AsyncGenerator[TestState, None]:
        yield ctx.state(TestState)
        with ctx.scope("nested", ctx.state(TestState).updated(value=value)):
            yield ctx.state(TestState)

    stream: AsyncIterator[TestState]
    async with ctx.scope("test", TestState(value=42)):
        elements: list[TestState] = []

        stream = ctx.stream(generator, 10)

    async for element in stream:
        elements.append(element)

    assert elements == [
        TestState(value=42),
        TestState(value=10),
    ]


@mark.asyncio
async def test_nested_streaming_streams_correctly():
    class TestState(State):
        value: int = 42
        other: str = "other"

    async def inner(value: int) -> AsyncGenerator[TestState, None]:
        yield ctx.state(TestState)
        with ctx.scope("inner", ctx.state(TestState).updated(value=value, other="inner")):
            yield ctx.state(TestState)

    async def outer(value: int) -> AsyncGenerator[TestState, None]:
        yield ctx.state(TestState)
        with ctx.scope("outer", ctx.state(TestState).updated(other="outer")):
            async for item in ctx.stream(inner, value):
                yield item

    stream: AsyncIterator[TestState]
    async with ctx.scope("test", TestState(value=42)):
        elements: list[TestState] = []

        stream = ctx.stream(outer, 10)

    async for element in stream:
        elements.append(element)

    assert elements == [
        TestState(value=42),
        TestState(value=42, other="outer"),
        TestState(value=10, other="inner"),
    ]


@mark.asyncio
async def test_streaming_context_completion_is_called_at_the_end_of_stream():
    completion_future = get_running_loop().create_future()

    class IterationMetric(State):
        value: int = 0

    metric: IterationMetric = IterationMetric()

    def completion(metrics: ScopeMetrics):
        nonlocal metric
        metric = metrics.metrics(merge=lambda current, nested: nested)[0]  # pyright: ignore[reportAssignmentType]
        completion_future.set_result(())

    async def generator(value: int) -> AsyncGenerator[int, None]:
        for i in range(0, value):
            ctx.record(
                IterationMetric(value=i),
                merge=lambda lhs, rhs: IterationMetric(value=lhs.value + rhs.value),
            )
            yield i

    stream: AsyncIterator[int]
    async with ctx.scope("test", completion=completion):
        elements: list[int] = []

        stream = ctx.stream(generator, 10)

    async for element in stream:
        elements.append(element)

    await completion_future
    assert metric.value == 45



from asyncio import (
    AbstractEventLoop,
    Future,
    gather,
    get_event_loop,
    iscoroutinefunction,
    run_coroutine_threadsafe,
)
from collections.abc import Callable, Coroutine
from contextvars import ContextVar, Token
from copy import copy
from itertools import chain
from logging import DEBUG, ERROR, INFO, WARNING, Logger, getLogger
from time import monotonic
from types import TracebackType
from typing import Any, Self, cast, final, overload
from uuid import uuid4

from haiway.state import State
from haiway.types import MISSING, Missing, not_missing
from haiway.utils import freeze

__all__ = [
    "ScopeMetrics",
    "MetricsContext",
]


@final
class ScopeMetrics:
    def __init__(
        self,
        *,
        trace_id: str | None,
        scope: str,
        logger: Logger | None,
        parent: Self | None,
        completion: Callable[[Self], Coroutine[None, None, None]] | Callable[[Self], None] | None,
    ) -> None:
        self.trace_id: str = trace_id or uuid4().hex
        self.identifier: str = uuid4().hex
        self.label: str = scope
        self._logger_prefix: str = (
            f"[{self.trace_id}] [{scope}] [{self.identifier}]"
            if scope
            else f"[{self.trace_id}] [{self.identifier}]"
        )
        self._logger: Logger = logger or getLogger(name=scope)
        self._parent: Self | None = parent if parent else None
        self._metrics: dict[type[State], State] = {}
        self._nested: set[ScopeMetrics] = set()
        self._timestamp: float = monotonic()
        self._finished: bool = False
        self._loop: AbstractEventLoop = get_event_loop()
        self._completed: Future[float] = self._loop.create_future()

        if parent := parent:
            parent._nested.add(self)

        freeze(self)

        if completion := completion:
            metrics: Self = self
            if iscoroutinefunction(completion):

                def callback(_: Future[float]) -> None:
                    run_coroutine_threadsafe(
                        completion(metrics),
                        metrics._loop,
                    )

            else:

                def callback(_: Future[float]) -> None:
                    completion(metrics)

            self._completed.add_done_callback(callback)

    def __del__(self) -> None:
        assert self.is_completed, "Deinitializing not completed scope metrics"  # nosec: B101

    def __str__(self) -> str:
        return f"{self.label}[{self.identifier}]@[{self.trace_id}]"

    def metrics(
        self,
        *,
        merge: Callable[[State | Missing, State], State | Missing] | None = None,
    ) -> list[State]:
        if not merge:
            return list(self._metrics.values())

        metrics: dict[type[State], State] = copy(self._metrics)
        for metric in chain.from_iterable(nested.metrics(merge=merge) for nested in self._nested):
            metric_type: type[State] = type(metric)
            merged: State | Missing = merge(
                metrics.get(  # current
                    metric_type,
                    MISSING,
                ),
                metric,  # received
            )

            if not_missing(merged):
                metrics[metric_type] = merged

        return list(metrics.values())

    @overload
    def read[Metric: State](
        self,
        metric: type[Metric],
        /,
    ) -> Metric | None: ...

    @overload
    def read[Metric: State](
        self,
        metric: type[Metric],
        /,
        default: Metric,
    ) -> Metric: ...

    def read[Metric: State](
        self,
        metric: type[Metric],
        /,
        default: Metric | None = None,
    ) -> Metric | None:
        return cast(Metric | None, self._metrics.get(metric, default))

    def record[Metric: State](
        self,
        metric: Metric,
        /,
        *,
        merge: Callable[[Metric, Metric], Metric] = lambda lhs, rhs: rhs,
    ) -> None:
        assert not self._completed.done(), "Can't record using completed metrics scope"  # nosec: B101
        metric_type: type[Metric] = type(metric)
        if current := self._metrics.get(metric_type):
            self._metrics[metric_type] = merge(cast(Metric, current), metric)

        else:
            self._metrics[metric_type] = metric

    @property
    def is_completed(self) -> bool:
        return self._completed.done() and all(nested.is_completed for nested in self._nested)

    @property
    def time(self) -> float:
        if self._completed.done():
            return self._completed.result()

        else:
            return monotonic() - self._timestamp

    async def wait(self) -> None:
        await gather(
            self._completed,
            *[nested.wait() for nested in self._nested],
            return_exceptions=False,
        )

    def _finish(self) -> None:
        assert (  # nosec: B101
            not self._completed.done()
        ), "Invalid state - called finish on already completed scope"

        assert (  # nosec: B101
            not self._finished
        ), "Invalid state - called completion on already finished scope"

        self._finished = True  # self is now finished

        self._complete_if_able()

    def _complete_if_able(self) -> None:
        assert (  # nosec: B101
            not self._completed.done()
        ), "Invalid state - called complete on already completed scope"

        if not self._finished:
            return  # wait for finishing self

        if any(not nested.is_completed for nested in self._nested):
            return  # wait for completing all nested scopes

        # set completion time
        self._completed.set_result(monotonic() - self._timestamp)

        # notify parent about completion
        if parent := self._parent:
            parent._complete_if_able()

    def log(
        self,
        level: int,
        message: str,
        /,
        *args: Any,
        exception: BaseException | None = None,
    ) -> None:
        self._logger.log(
            level,
            f"{self._logger_prefix} {message}",
            *args,
            exc_info=exception,
        )


@final
class MetricsContext:
    _context = ContextVar[ScopeMetrics]("MetricsContext")

    @classmethod
    def scope(
        cls,
        name: str,
        /,
        *,
        trace_id: str | None = None,
        logger: Logger | None = None,
        completion: Callable[[ScopeMetrics], Coroutine[None, None, None]]
        | Callable[[ScopeMetrics], None]
        | None,
    ) -> Self:
        current: ScopeMetrics
        try:  # check for current scope context
            current = cls._context.get()

        except LookupError:
            # create metrics scope when missing yet
            return cls(
                ScopeMetrics(
                    trace_id=trace_id,
                    scope=name,
                    logger=logger,
                    parent=None,
                    completion=completion,
                )
            )

        # or create nested metrics otherwise
        return cls(
            ScopeMetrics(
                trace_id=trace_id,
                scope=name,
                logger=logger or current._logger,  # pyright: ignore[reportPrivateUsage]
                parent=current,
                completion=completion,
            )
        )

    @classmethod
    def record[Metric: State](
        cls,
        metric: Metric,
        /,
        *,
        merge: Callable[[Metric, Metric], Metric] = lambda lhs, rhs: rhs,
    ) -> None:
        try:  # catch exceptions - we don't wan't to blow up on metrics
            cls._context.get().record(metric, merge=merge)

        except Exception as exc:
            cls.log_error(
                "Failed to record metric: %s",
                type(metric).__qualname__,
                exception=exc,
            )

    # - LOGS -

    @classmethod
    def log_error(
        cls,
        message: str,
        /,
        *args: Any,
        exception: BaseException | None = None,
    ) -> None:
        try:
            cls._context.get().log(
                ERROR,
                message,
                *args,
                exception=exception,
            )

        except LookupError:
            getLogger().log(
                ERROR,
                message,
                *args,
                exc_info=exception,
            )

    @classmethod
    def log_warning(
        cls,
        message: str,
        /,
        *args: Any,
        exception: Exception | None = None,
    ) -> None:
        try:
            cls._context.get().log(
                WARNING,
                message,
                *args,
                exception=exception,
            )

        except LookupError:
            getLogger().log(
                WARNING,
                message,
                *args,
                exc_info=exception,
            )

    @classmethod
    def log_info(
        cls,
        message: str,
        /,
        *args: Any,
    ) -> None:
        try:
            cls._context.get().log(
                INFO,
                message,
                *args,
            )

        except LookupError:
            getLogger().log(
                INFO,
                message,
                *args,
            )

    @classmethod
    def log_debug(
        cls,
        message: str,
        /,
        *args: Any,
        exception: Exception | None = None,
    ) -> None:
        try:
            cls._context.get().log(
                DEBUG,
                message,
                *args,
                exception=exception,
            )

        except LookupError:
            getLogger().log(
                DEBUG,
                message,
                *args,
                exc_info=exception,
            )

    def __init__(
        self,
        metrics: ScopeMetrics,
    ) -> None:
        self._metrics: ScopeMetrics = metrics
        self._token: Token[ScopeMetrics] | None = None

    def __enter__(self) -> None:
        assert (  # nosec: B101
            self._token is None and not self._metrics._finished  # pyright: ignore[reportPrivateUsage]
        ), "MetricsContext reentrance is not allowed"
        self._token = MetricsContext._context.set(self._metrics)

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        assert (  # nosec: B101
            self._token is not None
        ), "Unbalanced MetricsContext context enter/exit"
        MetricsContext._context.reset(self._token)
        self._metrics._finish()  # pyright: ignore[reportPrivateUsage]
        self._token = None



from haiway.state.attributes import AttributeAnnotation, attribute_annotations
from haiway.state.structure import State

__all__ = [
    "attribute_annotations",
    "AttributeAnnotation",
    "State",
]
